//  THeartMachine released to the public domain by the author
//  Adi L. Miller, adi@gomiller.com 26/01/2000
//
//  Excuse the dust. Spaghetti code, but it works
//
//  Please sumbit all bug-fixes, enhancments or additions to the author
//  or just Email me that you have downloaded it... -- adi@gomiller.com
//
//  See code for relevance info
//  See text past the End. for History
//
//  To Do:
//
//  1. Make the DrawSeries method more efficient for high-end use
//
//
unit uHeartMachine;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus;

const
  TheColors :Array[0..16] of TColor = (clRed, clLime, clYellow, clBlue, clFuchsia, clAqua, clLtGray, clDkGray, clWhite, clMaroon, clGreen, clOlive, clNavy, clPurple, clTeal, clGray, clSilver);

type
  TGraph = Array [1..100] of Integer; // Maximum number of points (640k should be enough for everyone...)

  TOnColorQuery = function (Sender: TObject; Index: Integer; AColor: TColor): TColor of object;
  THeartMachine = class(TGraphicControl)
  private
    FColorBackground,   // Color of component background
    FColorGrid: TColor; // Color of grid lines
    FHGridCount,        // Number of horizontal grid lines
    FHTickCount: Byte;  // Number of ticks visible in the component
    FMin,               // Minimum value of point
    FMax: Integer;      // Maximum value of point
    FSeries: Array of TGraph;  // Array of points (each for a serie)
    FSeriesVisible: Array of Boolean; // If each serie is visible or not
    FNumOfSeries: Integer;  // Number of series
    FPopupMenu: TPopupMenu;
    FMouseValue: Integer;
    FAppendValueToHint: Boolean;
    FOnColorQuery: TOnColorQuery; // A function like event (not sure if it's standard do
    Offscreen: TBitmap;     // Used for temporary screen canvas
    procedure SetColorBackground(const Value: TColor);
    procedure SetColorGrid(const Value: TColor);
    procedure SetHGridCount(const Value: Byte);
    function RealToGrid(I: Integer): Integer;
    function GridToReal(I: Integer): Integer;
    procedure SetHTickCount(const Value: Byte); // Translates the value of the point to the absolute location on the component canvas. Needs work!
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure AppShowHintEvent(var HintStr: string; var CanShow: Boolean; var HintInfo: THintInfo);
  protected
    procedure DrawBackground;
    procedure DrawGrid;
    procedure Paint; override;
    procedure DrawSeries;
    property PopupMenu;
    procedure PopupMenuDraw(Sender: TObject; ACanvas: TCanvas;
      ARect: TRect; Selected: Boolean);
    procedure PopupMeasureItem(Sender: TObject; ACanvas: TCanvas;
      var Width, Height: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddPoint(Series, Point: Integer; Name: String='');
    procedure PopupMenuClicked(Sender: TObject);
  published
    property Align;
    property ColorBackground: TColor read FColorBackground write SetColorBackground;
    property ColorGrid: TColor read FColorGrid write SetColorGrid;
    property HGridCount: Byte read FHGridCount write SetHGridCount;
    property Min: Integer read FMin write FMin;
    property Max: Integer read FMax write FMax;
    property HTickCount: Byte read FHTickCount write SetHTickCount;
    property Hint;
    property ShowHint;
    property AppendValueToHint: Boolean read FAppendValueToHint write FAppendValueToHint;
    property OnColorQuery: TOnColorQuery read FOnColorQuery write FOnColorQuery;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('System', [THeartMachine]);
end;

{ THeartMachine }

constructor THeartMachine.Create(AOwner: TComponent);
var
  I: Integer;
begin
  inherited;
  Offscreen := TBitmap.Create;
  ControlStyle := ControlStyle + [csOpaque];
  Application.OnShowHint := AppShowHintEvent;
  ShowHint := True;
  FPopupMenu := TPopupMenu.Create(Self);
  PopupMenu := FPopupMenu;
  Parent := TWinControl(AOwner);
  FNumOfSeries := 0;
  Width := 100;
  Height := 100;
  FMin := 1;
  FMax := 10;
  FHTickCount := 10;
  FHGridCount := 5;
  ColorBackground := clBlack;
  ColorGrid := clGreen;
  SetLength(FSeries, 1);
  SetLength(FSeriesVisible, 1);
  Align := alClient;
  For I := 1 to 99 do
    FSeries[0][I] := 0;
  FSeriesVisible[0] := True;
end;

destructor THeartMachine.Destroy;
begin
  Offscreen.Destroy;
  inherited;
end;

procedure THeartMachine.DrawBackground;
var
  tmpRect: TRect;
begin
  tmpRect.Top := 0;
  tmpRect.Left := 0;
  tmpRect.Bottom := Height;
  tmpRect.Right := Width;
  OffScreen.Canvas.Brush.Color := ColorBackground;
  OffScreen.Canvas.Brush.Style := bsSolid;
  OffScreen.Canvas.FillRect(tmpRect);
end;

procedure THeartMachine.DrawGrid;
var
  I, II: Integer;
  Factor, XFactor: Real;
begin
  DrawBackground;

  // Horizontal grid lines
  Factor := ((Height-1) / (HGridCount-1));
  OffScreen.Canvas.Pen.Color := ColorGrid;
  OffScreen.Canvas.Pen.Width := 1;
  For I := 0 to HGridCount-1 do
  begin
    OffScreen.Canvas.MoveTo(0,Trunc(I*Factor));
    OffScreen.Canvas.LineTo(Width,Trunc(I*Factor));
  end;

  // Vertical grid lines
  XFactor := Width / HTickCount;
  For II := 100 downto 100-HTickCount do
  begin
    OffScreen.Canvas.MoveTo(Trunc(XFactor*(HTickCount-(100-II))), 0);
    OffScreen.Canvas.LineTo(Trunc(XFactor*(HTickCount-(100-II))), Height);
  end;
end;

procedure THeartMachine.AddPoint(Series, Point: Integer; Name: String='');
var
  I,i1: Integer;
  tmpMenuItem: TMenuItem;
  MaxMax:Longint;
begin
  // Making a simple range check
  if (Point > Max) then Max:=Point;
  if (Point < Min) then Point:=Min;
//    raise ERangeError.Create('Point value out of range ('+IntToStr(Point)+')');
  // Checking to see if a creation of a new serie is in order if so create
  if FNumOfSeries < Series then
  begin
    Inc(FNumOfSeries);
    SetLength(FSeriesVisible, Series+1);
    SetLength(FSeries, Series+1);    // Assuming that AddPoint does
    For I := 1 to 100 do             // address a serie higher then
      FSeries[Series][I] := 0;       // one of an already existing serie
    FSeriesVisible[Series] := True;  // Potential bug/mess-up
  end;
  // Creat a new menu item if missing for current serie. Should happen:
  //    1. When a new serie was just added at the section above
  //    2. When the first AddPoint called, cause .Create didn't creat the first menu itme
  if PopupMenu.Items.Count < Series+1 then
  begin
    tmpMenuItem := TMenuItem.Create(self);
    tmpMenuItem.Caption := Name;
    tmpMenuItem.Tag := Series;
    tmpMenuItem.OnClick := PopupMenuClicked;
    tmpMenuItem.Checked := True;
    tmpMenuItem.OnDrawItem := PopupMenuDraw;
    tmpMenuItem.OnMeasureItem := PopupMeasureItem;
    PopupMenu.Items.Add(tmpMenuItem);
    PopupMenu.OwnerDraw := True;
  end;
  // Update the name of the serie
  PopupMenu.Items[Series].Caption := Name;
  // Moving all points back
  For I := 2 to 100 do begin
    FSeries[Series][I-1] := FSeries[Series][I];
  end;
  // Placing Point in place 100
  FSeries[Series][100] := Point;

  MaxMax:=Min;
  For I := 1 to 100 do
   For I1:=0 to FNumOfSeries do
    if FSeries[i1][I]>MaxMax then MaxMax:=FSeries[i1][I];
  Max:=MaxMax+round((MaxMax*0.1));

  Refresh;
end;

procedure THeartMachine.DrawSeries;
var
  I, II: Integer;
  XFactor: Real;
begin
  DrawGrid;
  OffScreen.Canvas.Pen.Width := 1;
  XFactor := Width / HTickCount;
  For I := 0 to Length(FSeries) - 1 do
    if FSeriesVisible[I] then
    begin
      if Assigned(FOnColorQuery) then
        OffScreen.Canvas.Pen.Color := FOnColorQuery(Self, I, TheColors[I])
      else
        OffScreen.Canvas.Pen.Color := TheColors[I];
      OffScreen.Canvas.MoveTo(Trunc(XFactor*(HTickCount)), RealToGrid(FSeries[I][100]));
      For II := 100 downto 101-HTickCount do
        OffScreen.Canvas.LineTo(Trunc(XFactor*(HTickCount-(100-II))), RealToGrid(FSeries[I][II]));
    end;
end;

function THeartMachine.RealToGrid(I: Integer): Integer;
begin
  // Translates the value given by the "user" to a value absolute to
  // the component's canvas.
  Result := Trunc((Height-1)-((I-Min)*(((Height-1)/(Max-Min)))));
end;

function THeartMachine.GridToReal(I: Integer): Integer;
begin
  Result := (Max+Min)-Trunc(((I / (Height-1))*(Max-Min))+Min);
end;

procedure THeartMachine.Paint;
begin
  OffScreen.Width := Width;
  OffScreen.Height := Height;
  DrawSeries;
  Canvas.Draw(0, 0, OffScreen);
end;

procedure THeartMachine.AppShowHintEvent(var HintStr: string;
  var CanShow: Boolean; var HintInfo: THintInfo);
begin
  HintInfo.ReshowTimeout := 10;
  if AppendValueToHint then
    if HintStr = '' then
      HintStr := 'Value='+IntToStr(FMouseValue)
    else
      HintStr := Hint + '(Value='+IntToStr(FMouseValue)+')';
  CanShow := True;
end;

procedure THeartMachine.WMMouseMove(var Message: TWMMouseMove);
begin
  if ShowHint and AppendValueToHint then
    FMouseValue := GridToReal(Message.YPos);
end;

// Popup menu

procedure THeartMachine.PopupMeasureItem(Sender: TObject; ACanvas: TCanvas;
  var Width, Height: Integer);
begin
  // To give menu items broader boundry
  Width := ACanvas.TextWidth((Sender as TMenuItem).Caption)+10;
end;

procedure THeartMachine.PopupMenuClicked(Sender: TObject);
var
  B: Boolean;
  tmpMenuItem: TMenuItem;
begin
  // To make a serie disappear/reappear from the graph
  tmpMenuItem := (Sender as TMenuItem);
  B := not FSeriesVisible[tmpMenuItem.Tag];
  FSeriesVisible[tmpMenuItem.Tag] := B;
  tmpMenuItem.Checked := B;
  DrawSeries;
end;

procedure THeartMachine.PopupMenuDraw(Sender: TObject; ACanvas: TCanvas;
  ARect: TRect; Selected: Boolean);
var
  AMenuItem: TMenuItem;
  tmpRect: TRect;
begin
  // Draw the menu items
  AMenuItem := (Sender as TMenuItem);
  if Selected then
  begin
    // Highlighted rect
    ACanvas.Brush.Color := clActiveCaption;
    ACanvas.FillRect(ARect);
  end else
  begin
    // Un-highlighted rect
    ACanvas.Brush.Color := clBtnFace;
    ACanvas.FillRect(ARect);
  end;
  // Draw menu text with the assigned color from the TheColors set
  if Assigned(FOnColorQuery) then
    ACanvas.Font.Color := FOnColorQuery(Self, AMenuItem.Tag, TheColors[AMenuItem.Tag])
  else
    ACanvas.Font.Color := TheColors[AMenuItem.Tag];
  ACanvas.TextOut(ARect.Left+12, ARect.Top+3, AMenuItem.Caption);
  // Draw V mark
  if AMenuItem.Checked then
  begin
    if Selected then
      ACanvas.Pen.Color := clCaptionText
    else
      ACanvas.Pen.Color := clMenuText;
    ACanvas.Pen.Width := 2;
    ACanvas.MoveTo(ARect.Left+1, ARect.Top+9);
    ACanvas.LineTo(ARect.Left+4, ARect.Top+12);
    ACanvas.LineTo(ARect.Left+9, ARect.Top+6);
  // Or use this code instead of the drawn V mark for a simple square
{    ACanvas.Brush.Color := clBlack;
    with tmpRect do
    begin
      Left := 3; Top := ARect.Top+8; Right := 7; Bottom := ARect.Top+12;
    end;
    ACanvas.FillRect(tmpRect);}
  end;
end;

// Get and Set

procedure THeartMachine.SetColorBackground(const Value: TColor);
begin
  // rtfm
  if FColorBackground <> Value then
  begin
    FColorBackground := Value;
    Invalidate;
  end;
end;

procedure THeartMachine.SetColorGrid(const Value: TColor);
begin
  // rtfm
  if FColorGrid <> Value then
  begin
    FColorGrid := Value;
    Invalidate;
  end;
end;

procedure THeartMachine.SetHGridCount(const Value: Byte);
begin
  // rtfm
  if FHGridCount <> Value then
  begin
    FHGridCount := Value;
    Invalidate;
  end;
end;

procedure THeartMachine.SetHTickCount(const Value: Byte);
begin
  // rtfm
  if FHTickCount <> Value then
  begin
    FHTickCount := Value;
    Invalidate;
  end;
end;

end.

// You made it...
//
// History:

// v0.03 (30/01/2000)
//
// Fixed component flickering problem, thanks to Kambiz

// v0.02
//
// Changed every div to a / for accuracy, fixed DrawGrid and RealToGrid accuracy
// Added the OnColorQuery event as a function
// Added tooltip with the Y value

